#ifndef __MinimalThreadAbstraction_h__
#define __MinimalThreadAbstraction_h__

#include "AbstractionFunction.h"
#include "ThreadStateGraph.h"

class MinimalThreadAbstraction
{
public:
    static void abstract( ThreadStateGraph &input, ThreadStateGraph &output, AbstractionFunction &abstractionFunction )
    {
        MinimalThreadAbstraction( input, output, abstractionFunction ).run();
    }

private:
    ThreadStateGraph *_input;
    ThreadStateGraph *_output;
    AbstractionFunction *_abstractionFunction;

    MinimalThreadAbstraction( ThreadStateGraph &input, ThreadStateGraph &output, AbstractionFunction &abstractionFunction ) :
        _input( &input ), _output( &output ), _abstractionFunction( &abstractionFunction )
    {
    }

    void run()
    {
        initalizeOutputGraph();

        createAllStates();
        createAllTransitions();
    }

    void createAllStates()
    {
        Map<String, ThreadState>::iterator it;

        for ( it = _input->_graph.begin(); it != _input->_graph.end(); ++it )
        {
            handleSingleInputState( it->first );
        }
    }

    String inputLabelToOutputLabel( String &inputLabel )
    {
        if ( _input->_initialState == inputLabel )
        {
            return "Init";
        }
        else
        {
            return _input->_graph[inputLabel]._atomicPropositions.toString();
        }
    }

    void handleSingleInputState( String inputLabel )
    {
        String outputLabel = inputLabelToOutputLabel( inputLabel );
        _abstractionFunction->add( inputLabel, outputLabel );

        if ( !_output->_graph.exists( outputLabel ) )
        {
            createNewState( inputLabel, outputLabel );
        }
        else
        {
            adjustExistingState( inputLabel, outputLabel );
        }
    }

    void createNewState( String inputLabel, String outputLabel )
    {
        ThreadState state;

        state._label = outputLabel;

        state._requestedEvents = _input->_graph[inputLabel]._requestedEvents;
        state._blockedEvents = _input->_graph[inputLabel]._blockedEvents;
        state._atomicPropositions = _input->_graph[inputLabel]._atomicPropositions;

        _output->_graph[state._label] = state;
    }

    void adjustExistingState( String inputLabel, String outputLabel )
    {
        ThreadState &inputState( _input->_graph[inputLabel] );
        ThreadState &outputState( _output->_graph[outputLabel] );

        for ( unsigned i = 0; i < inputState._requestedEvents.size(); ++i )
        {
            if ( !outputState._requestedEvents.exists( inputState._requestedEvents[i] ) )
            {
                outputState._requestedEvents.append( inputState._requestedEvents[i] );
            }
        }

        Vector<Event> copyOfBlocked = outputState._blockedEvents;

        for ( unsigned i = 0; i < copyOfBlocked.size(); ++i )
        {
            if ( !inputState._blockedEvents.exists( copyOfBlocked[i] ) )
            {
                outputState._blockedEvents.eraseByValue( copyOfBlocked[i] );
            }
        }
    }

    void createAllTransitions()
    {
        Map<String, ThreadState>::iterator it;

        for ( it = _input->_graph.begin(); it != _input->_graph.end(); ++it )
        {
            handleTransitionsOfInputState( it->first );
        }
    }

    void handleTransitionsOfInputState( String inputLabel )
    {
        ThreadState &inputState( _input->_graph[inputLabel] );

        for ( unsigned i = 0; i < inputState._successors.size(); ++i )
        {
            handleSingleTransition( inputState, inputState._successors[i] );
        }
    }

    void handleSingleTransition( ThreadState &inputState, Transition &transition )
    {
        ThreadState &outputState( _output->_graph[inputLabelToOutputLabel( inputState._label )] );
        ThreadState &outputSuccessor( _output->_graph[inputLabelToOutputLabel( transition._state )] );

        Transition successorTransition( transition._event, outputSuccessor._label );
        Transition predecessorTransition( transition._event, outputState._label );

        if ( !outputState._successors.exists( successorTransition ) )
        {
            outputState._successors.append( successorTransition );
            outputSuccessor._predecessors.append( predecessorTransition );
        }
    }

    void initalizeOutputGraph()
    {
        _output->_events = _input->_events;
        _output->_atomicPropositions = _input->_atomicPropositions;
        _output->_initialState = inputLabelToOutputLabel( _input->_initialState );
    }
};

#endif // __MinimalThreadAbstraction_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
