#ifndef __SequenceMaster_h__
#define __SequenceMaster_h__

#include "BThread.h"
#include "IEvent.h"
#include "SequenceNumberPayload.h"

class SequenceMaster : public BThread
{
public:
    enum {
        INITIAL_SEQUENCE_NUMBER = 0,
    };

    SequenceMaster() : _nextSequenceNumber( INITIAL_SEQUENCE_NUMBER )
    {
    }

    void entryPoint()
    {
        while ( true )
        {
            waitForSequenceNumberRequest();
            sendSequenceNumber();
            increaseSequenceNumber();
        }
    }

private:
    unsigned _nextSequenceNumber;
    unsigned _sizeOfNextPacket;
    AtomicPropositions _atomicPropositions;

    Vector<Event> allRequestTypes()
    {
        Vector<Event> types;

        types.append( IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER );
        types.append( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER );

        return types;
    }

    Vector<Event> allResponseTypes()
    {
        Vector<Event> types;

        types.append( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
        types.append( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );

        return types;
    }

    void waitForSequenceNumberRequest()
    {
        clearEventVectors();

        _watched = allRequestTypes();
        _blocked = allResponseTypes();

        bSync( _requested, _watched, _blocked, "Waiting" );

        _sizeOfNextPacket = SequenceNumberPayload( lastEvent().payload() ).getBytesToBeSent();
    }

    void sendSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
        };

        clearEventVectors();
        blockAllRelevantEvents();

        SequenceNumberPayload payload( _nextSequenceNumber, DONT_CARE );
        String nextLabel;

        switch ( lastEvent().type() )
        {
        case IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( Event( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER, ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER );
            nextLabel = "HandleAckSender";
            break;

        case IEvent::SYN_HANDLER_REQUEST_SEQUENCE_NUMBER:
            _requested.append( Event( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER, ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_SYN_HANDLER );
            nextLabel = "HandleSynHandler";
            break;

        case IEvent::FIN_HANDLER_REQUEST_SEQUENCE_NUMBER:
            nextLabel = "HandleFinHandler";
            _requested.append( Event( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER, ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_FIN_HANDLER );
            break;

        case IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER:
            nextLabel = "HandlePushSender";
            _requested.append( Event( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER, ConstSimpleData( &payload, sizeof(payload) ) ) );
            _blocked.eraseByValue( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );
            break;

        default:
            break;
        }

        _atomicPropositions.clear();
        _atomicPropositions.add( "SeqMasterReplying" );

        bSync( _requested, _watched, _blocked, _atomicPropositions, nextLabel );
    }

    void blockAllRelevantEvents()
    {
        _blocked = allRequestTypes();
        Vector<Event> responses = allResponseTypes();
        _blocked += responses;
    }

    void increaseSequenceNumber()
    {
        _nextSequenceNumber += _sizeOfNextPacket;
    }
};

#endif // __SequenceMaster_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
