#ifndef __TcpPushSender_h__
#define __TcpPushSender_h__

#include "BThread.h"
#include "IEvent.h"
#include "TargetPortPayload.h"

class TcpPushSender : public BThread
{
public:
    void entryPoint()
    {
        while ( true )
        {
            waitForPushesAndAnnouncements();

            switch ( lastEvent().type() )
            {
            case IEvent::TCP_SEND_PUSH:
                handlePushRequest();
                break;
            case IEvent::TCP_ANNOUNCE_TARGET_PORT:
                handleTargetPort();
                break;
            default:
                break;
            }
        }
    }

private:
    unsigned _sequenceNumber;
    HeapData _pushBuffer;
    unsigned _targetPort;

    void handlePushRequest()
    {
        storePushBuffer();
        getFreshSequenceNumber();
        sendIpPacket();
    }

    void handleTargetPort()
    {
        _targetPort = TargetPortPayload( lastEvent().payload() ).getTargetPort();
    }

    void waitForPushesAndAnnouncements()
    {
        clearEventVectors();
        _watched.append( IEvent::TCP_SEND_PUSH );
        _watched.append( IEvent::TCP_ANNOUNCE_TARGET_PORT );

        _blocked.append( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER );
        _blocked.append( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );

        bSync( _requested, _watched, _blocked, "Wait" );
    }

    void storePushBuffer()
    {
        _pushBuffer = lastEvent().payload();
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, _pushBuffer.size() );

        _requested.append( Event( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER, ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, "ReqSequence" );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );
        bSync( _requested, _watched, _blocked, "GetSequence" );

        _sequenceNumber = SequenceNumberPayload( lastEvent().payload() ).getSequenceNumber();
    }

    void sendIpPacket()
    {
        TcpSegment push;
        preparePush( push );
        sendPush( push );
    }

    void preparePush( TcpSegment &push )
    {
        push._payload = _pushBuffer;

        push.header().setSourcePort( ServerConfiguration::LISTENING_PORT );
        push.header().setDestinationPort( _targetPort );

        push.header().setAckNumber( 0 );
        push.header().setSequenceNumber( _sequenceNumber );

        push.header().setPush();

        push.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendPush( TcpSegment &push )
    {
        HeapData serialized;
        push.serialize( serialized );

        clearEventVectors();
        _requested.append( Event( IEvent::TCP_SEND_DATA, serialized ) );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SEND_DATA );
        bSync( _requested, _watched, _blocked, "Push" );
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_SEND_PUSH ) result.append( IEvent::TCP_SEND_PUSH );
        if ( event != IEvent::TCP_ANNOUNCE_TARGET_PORT ) result.append( IEvent::TCP_ANNOUNCE_TARGET_PORT );
        if ( event != IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );

        return result;
    }
};

#endif // __TcpPushSender_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
