#include <cxxtest/TestSuite.h>

#include "AbstractSafetyPatcher.h"
#include "AbstractionData.h"
#include "AbstractionFunction.h"
#include "BridgeToHttpSpanner.h"
#include "BridgeToTcpSpanner.h"
#include "CegarRefiner.h"
#include "HttpDispatcherSimulatorSpanner.h"
#include "HttpErrorHandlerSpanner.h"
#include "HttpFoundHandlerSpanner.h"
#include "HttpGetSimulatorSpanner.h"
#include "IEvent.h"
#include "MinimalThreadAbstraction.h"
#include "MockErrno.h"
#include "MockForRawSocket.h"
#include "MultipleThreadIntersector.h"
#include "NFA.h"
#include "RawSocketSimulatorSpanner.h"
#include "RealMalloc.h"
#include "SafetyModelChecker.h"
#include "SafetyPatch.h"
#include "SequenceMasterSpanner.h"
#include "TcpAckSenderSpanner.h"
#include "TcpDispatcherSimulatorSpanner.h"
#include "TcpFinHandlerSpanner.h"
#include "TcpPushHandlerSpanner.h"
#include "TcpPushSenderSpanner.h"
#include "TcpSynHandlerSpanner.h"
#include "ThreadPartitioner.h"
#include "ThreadStateGraph.h"

#include "safety/SynBeforeFin.h"

class MockForSpan :
    public MockForRawSocket
{
public:
};

class SpanTestSuite : public CxxTest::TestSuite
{
public:
    MockForSpan mock;

    ThreadStateGraph _sequenceMaster;
    ThreadStateGraph _tcpDispatcherSimulator;
    ThreadStateGraph _tcpSynHandler;
    ThreadStateGraph _tcpPushHandler;
    ThreadStateGraph _tcpFinHandler;
    ThreadStateGraph _tcpAckSender;
    ThreadStateGraph _tcpPushSender;

    ThreadStateGraph _bridgeToTcp;
    ThreadStateGraph _bridgeToHttp;

    ThreadStateGraph _httpDispatcherSimulator;
    ThreadStateGraph _httpErrorHandler;
    ThreadStateGraph _httpFoundHandler;

    ThreadStateGraph _rawSocketSimulator;
    ThreadStateGraph _httpGetSimulator;

    ThreadStateGraph _fullProgram;

    AbstractionData _abstractionData;

    void test_program_properties()
    {
        spanThreadGraphs();

        // createFullProgramStateGraph();

        createAbstractGraph();
        cegarSafetyTest();
        // patch();
    }

    void spanThreadGraphs()
    {
        TS_TRACE( "\tSpanning thread graphs" );

        SequenceMasterSpanner().span( _sequenceMaster );
        TcpDispatcherSimulatorSpanner().span( _tcpDispatcherSimulator );
        TcpSynHandlerSpanner().span( _tcpSynHandler );
        TcpPushHandlerSpanner().span( _tcpPushHandler );
        TcpFinHandlerSpanner().span( _tcpFinHandler );
        TcpAckSenderSpanner().span( _tcpAckSender );
        TcpPushSenderSpanner().span( _tcpPushSender );

        BridgeToTcpSpanner().span( _bridgeToTcp );
        BridgeToHttpSpanner().span( _bridgeToHttp );

        HttpDispatcherSimulatorSpanner().span( _httpDispatcherSimulator );
        HttpErrorHandlerSpanner().span( _httpErrorHandler );
        HttpFoundHandlerSpanner().span( _httpFoundHandler );

        RawSocketSimulatorSpanner().span( _rawSocketSimulator );
        HttpGetSimulatorSpanner().span( _httpGetSimulator );
    }

    void createFullProgramStateGraph()
    {
        MultipleThreadIntersector intersector;

        intersector.addThread( &_sequenceMaster );
        intersector.addThread( &_tcpDispatcherSimulator );
        intersector.addThread( &_tcpSynHandler );
        intersector.addThread( &_tcpPushHandler );
        intersector.addThread( &_tcpFinHandler );
        intersector.addThread( &_tcpAckSender );
        intersector.addThread( &_tcpPushSender );
        intersector.addThread( &_bridgeToHttp );
        intersector.addThread( &_bridgeToTcp );
        intersector.addThread( &_httpDispatcherSimulator );
        intersector.addThread( &_httpErrorHandler );
        intersector.addThread( &_httpFoundHandler );
        intersector.addThread( &_rawSocketSimulator );
        intersector.addThread( &_httpGetSimulator );

        intersector.intersect( _fullProgram, ' ' );

        TS_TRACE( "Before applying axioms:" );
        TS_TRACE( _fullProgram._graph.size() );

        TS_TRACE( "After applying axioms:" );
        _fullProgram.applyAxioms();
        TS_TRACE( _fullProgram._graph.size() );
    }

    void createAbstractGraph()
    {
        partitionToModules();
        abstractEachModule();
        composeAbstractModules();

        _abstractionData._composedGraph->dump();
    }

    Vector<BModule> _bModules;

    void partitionToModules()
    {
        Vector<ThreadStateGraph *> threads;

        threads.append( &_sequenceMaster );
        threads.append( &_tcpDispatcherSimulator );
        threads.append( &_tcpSynHandler );
        threads.append( &_tcpPushHandler );
        threads.append( &_tcpFinHandler );
        threads.append( &_tcpAckSender );
        threads.append( &_tcpPushSender );
        threads.append( &_bridgeToHttp );
        threads.append( &_bridgeToTcp );
        threads.append( &_httpDispatcherSimulator );
        threads.append( &_httpErrorHandler );
        threads.append( &_httpFoundHandler );
        threads.append( &_rawSocketSimulator );
        threads.append( &_httpGetSimulator );

        _bModules = ThreadPartitioner::partition( threads );

        for ( unsigned i = 0; i < _bModules.size(); ++i )
        {
            printf( "Module %u: ", i );

            for ( unsigned j = 0; j < _bModules[i]._threads.size(); ++j )
            {
                printf( "%s ", _bModules[i]._threads[j]->_threadName.ascii() );
            }

            printf( "\n" );
        }
    }

    void abstractEachModule()
    {
        for ( unsigned i = 0; i < _bModules.size(); ++i )
        {
            TS_TRACE( Stringf( "Abstracting module #%u", i ).ascii() );
            abstractSingleModule( _bModules[i] );
        }
    }

    void abstractSingleModule( BModule &module )
    {
        char moduleInnerSeparationChar = '.';

        MultipleThreadIntersector intersector;

        for ( unsigned i = 0; i < module._threads.size(); ++i )
        {
            intersector.addThread( module._threads[i] );
        }

        ThreadStateGraph *composedModule = new ThreadStateGraph;
        intersector.intersect( *composedModule, moduleInnerSeparationChar );
        TS_TRACE( Stringf( "Size of module before abstraction: %u", composedModule->_graph.size() ).ascii() );

        ThreadStateGraph *abstractModule = new ThreadStateGraph;
        AbstractionFunction *abstractionFunction = new AbstractionFunction;

        MinimalThreadAbstraction::abstract( *composedModule, *abstractModule, *abstractionFunction );
        TS_TRACE( Stringf( "Size of module after abstraction: %u", abstractModule->_graph.size() ).ascii() );

        _abstractionData._abstractModules.append( AbstractionData::ModuleData( composedModule, abstractModule, abstractionFunction ) );
    }

    void composeAbstractModules()
    {
        _abstractionData._composedGraph = new ThreadStateGraph;

        char separationCharBetweenModules = ' ';
        MultipleThreadIntersector intersector;

        for ( unsigned i = 0; i < _abstractionData._abstractModules.size(); ++i )
        {
            intersector.addThread( _abstractionData._abstractModules[i]._abstract );
        }

        intersector.intersect( *_abstractionData._composedGraph, separationCharBetweenModules );

        TS_TRACE( "Abstract program Before applying axioms:" );
        TS_TRACE( _abstractionData._composedGraph->_graph.size() );

        TS_TRACE( "After applying axioms:" );
        _abstractionData._composedGraph->applyAxioms();
        TS_TRACE( _abstractionData._composedGraph->_graph.size() );
    }

    void cegarSafetyTest()
    {
        NFA safetyProperty;

        safetyProperty = SynBeforeFin::getNfa();

        NFA productAutomaton;
        NFA::Run counterExample;
        DirectAbstraction directAbstraction( productAutomaton, _abstractionData );

        bool flag = false;

        while ( !flag )
        {
            counterExample.clear();

            if ( SafetyModelChecker::checkSafetyProperty( *_abstractionData._composedGraph, safetyProperty, counterExample, productAutomaton ) )
            {
                TS_TRACE( "Property Holds!" );
                flag = true;
            }
            else
            {
                TS_TRACE( "Property doesn't hold on the abstract graph" );

                dumpCounterExample( counterExample );

                CegarRefiner cegarRefiner;

                if ( !cegarRefiner.checkAndRefine( productAutomaton, counterExample, _abstractionData, directAbstraction ) )
                {
                    printf( "The counter example is genuine!\n" );
                    flag = true;
                }
            }
        }
    }

    void dumpCounterExample( NFA::Run &counterExample )
    {
        printf( "Counter example:\n" );
        for ( unsigned i = 0; i < counterExample._states.size() - 1; ++i )
        {
            printf( "%s\n\t(%s)\n", counterExample._states[i].ascii(), counterExample._letters[i].ascii() );
        }
        printf( "%s\n", counterExample._states.last().ascii() );
    }

    void patch()
    {
        NFA safetyProperty;

        safetyProperty = SynBeforeFin::getNfa();

        NFA::Run counterExample;
        NFA productAutomaton;

        try
        {
            if ( SafetyModelChecker::checkSafetyProperty( *_abstractionData._composedGraph, safetyProperty, counterExample, productAutomaton ) )
            {
                TS_TRACE( "Property holds!" );
                return;
            }
            else
            {
                TS_TRACE( "Property doesnt hold!" );
            }

            SafetyPatch patch;

            patch = AbstractSafetyPatcher::patch( productAutomaton, _abstractionData );
            patch.dump();
        }
        catch ( const Error &e )
        {
            TS_TRACE( Stringf( "Patching failed! Error: %u", e.code() ).ascii() );
        }
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
